/*
 *
 *
 *    CwModem.ino
 *
 *    A simple terminal for CW.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

//
//  Set CW_DATA_ANALOG to 'true' to read analog AFSK data from A0.  This
//  will disable all other analog pins, but digital pins will still be
//  available
//
//  Set to 'false' if using an external tone detector such as the LM567,
//  or NJM2211, or another PLL-based circuit.
//
#define CW_DATA_ANALOG false

#if CW_DATA_ANALOG == true
	//
	//  If reading AFSK data from A0...
	//

	//
	//  Define the center frequency (Hz) of the tone tetector.
	//
	#define CW_FREQ (700.0)

	//  Set the sampling divisor to determine the sample rate; divisor
	//  should be set to one of the following values, depending on the
	//  sample rate desired.  It is STRONGLY recommended to use the 
	//  default value (128).  This will give a sample rate of ~9.6 kSps.
	//
	//  The input audio should be low-pass filtered with a cutoff
	//  below the Nyquist frequency for the selected sample rate.
	//  The typical 3kHz filter in a typical HF SSB receiver is
	//  sufficient for most purposes.  Narrower filters will result
	//  in better performance.
	//
	//  Divisor   ->    Sample Rate   ->   Nyquist Frequency
	//    128            9615.4 S/s            4.81 kHz
	//     64           19230.8 S/s            9.62 kHz
	//     32           38461.5 S/s           19.23 kHz
	//
	#define SAMPLE_RATE_DIVISOR (128)

	//
	//  Define the zero-volts value to be read from the A/D.  If your
	//  hardware has a perfectly centered DC offset (using a high-
	//  quality voltage divider, etc.), just leave this at 127.  If
	//  you have a larger DC offset, you can change this *slightly*
	//  to compensate.
	//
	#define ANALOG_DC_OFFSET (127)

	//
	//  Set the CPU frequency, in Hz.  Use 16000000.0 for an Arduino
	//  UNO or Mega; some Pro boards use 8000000.0.
	//
	#define CPU_FREQ (16000000.0)
#else
	//
	//  If reading digital detector voltage from detector IC...
	//

	//
	//  REQUIRED: CW RX detector input (digital input)
	//
	//  This is the pin attached to the detector output line.  The logic
	//  assumes that the decoder IC produces an open-collector output from
	//  a pin that is tied to V+ through a pull-up resistor.  That is, a
	//  detected tone is signaled by this pin being pulled LOW by the
	//  detector IC.
	//
	#define CW_RX_PIN (A0)
#endif

//
//  OPTIONAL: PTT output (digital output)
//
//  This is the pin used to enable the radio transmitter.  If used, this
//  needs to be connected to some kind of output switching device, like an
//  n-channel FET, NPN BJT, or solid state relay, which is then connected
//  to the PTT input on the transmitter.  This pin will be driven HIGH to
//  enable the transmitter.
//
//  This pin can also be attached to a current-limited LED as a transmit
//  PTT indicator.
//
#define CW_PTT_PIN (A1)

//
//  REQUIRED: CW TX (key) output (digital output)
//
//  This is the pin used to transmit key pulses. This needs to be connected
//  to some kind of output switching device, like an n-channel FET, NPN BJT,
//  or solid state relay, which is then connected to the KEY input on the
//  transmitter.  This pin will be driven HIGH to generate key closures.
//
//  This pin can also be attached to a current-limited LED as a transmit
//  key-down indicator.
//
#define CW_KEY_PIN (A2)

//
//  OPTIONAL: Command LED (digital output)
//
//  If attached to an LED, the LED will blink momentarily when data is
//  received from the host.
//
//  This is disabled by default, since most boards have LEDs attached to
//  the serial port.  Set to something other than zero to drive a custom
//  LED attached to another port.
//
#define CMD_LED (0)

//
//  OPTIONAL: CW detector LED
//
//  If attached to an LED, the LED will blink whenever tone is detected,
//  either by the external decoder IC, or by the software detector.
//
//  Set to zero to disable.  Leave this set to 13 to blink the built-in
//  LED attached to D13 on most boards.
//
#define CW_LED_PIN (13)

//
//  REQUIRED: Host Serial Port
//
//  Assign a specific hardware port to be the interface to the host;
//  by default, the USB connection on most boards is 'Serial', which
//  is also the hardware TTL serial port on the UNO board.  The hardware
//  serial device on the Leonardo is Serial1.
//
//  Leave this alone unless you have a Leonardo, or want to use a
//  non-default host port on a Mega board.
//
#define HostSerial Serial

//
//  REQUIRED: host port serial rate
//
//  Set the 'baud rate' of the serial port.
//
#define SerialRate (9600)

// ================== END OF CONFIGURATION SETTINGS ==================

// local headers
#include "CircularBuffer.h"
#include "CwTimingLogic.h"
#include "CwDecoderLogic.h"
#include "SplitCommand.h"
#include "Elapsed.h"
#include "Bounce2.h"

#if CW_DATA_ANALOG == true
#include "avraudio.h"
#include "favg.h"
#include "tone.h"
// set the short-term envelope integrator window; do not change this
//    unless you *absolutely* know what you are doing.
#define STA (5)
#endif

// the version string
#define VERSION_STRING ("CwModem 1.0 (beta-20161028a)")

// CR+LF
#define LOCAL_CRLF ("\r\n")

// import some namespaces
using namespace KK5JY::Collections;
using namespace KK5JY::CW;

// buffer for element timing data
static CircularBuffer<CwElement> CwBuffer(32);

// buffer for decoded elements
static CircularBuffer<DataElement> ElementBuffer(32);

// a TX buffer
static CircularBuffer<char> TxBuffer(32);

// clock for CW
static CwTimingLogic CwTiming;

// decoder for CW
static CwDecoderLogic CwDecoder;

// the clock restoration logic
static TimingLogic *Timing = &CwTiming;

// the decoder
static DecoderLogic *Decoder = &CwDecoder;

#if CW_DATA_ANALOG == true
// RX CW state - start with LOW (unkeyed or no-tone)
static bool lastState = LOW;

// audio detector output state
static bool audioState = LOW;

// define a convenience identifier that fully qualifies the detector type
#define \
	Detector \
	KK5JY::DSP::ToneDetector<int16_t, KK5JY::DSP::FixedAverager<int32_t, int32_t>, KK5JY::DSP::FixedPoint<int16_t> >

// allocate the detector object
static Detector detector(Detector::Config::Create(CW_FREQ, GetFs(CPU_FREQ, SAMPLE_RATE_DIVISOR)));
#else
// debounce for the key input digital pin
static Bounce KeyInput;

// RX CW state - start with 'true' (unkeyed or no-tone)
static bool lastState = true;
#endif

// a string to hold incoming data
static String inputString; // start with Version, WPM queries

// and another for command data
static String cmdText = "";

// command parsing
static String command = "";
static String argument = "";

// whether the string is complete
static boolean inputReady = false;

// I/O buffer
static char ioBuffer[32];

// the last RX data from the host
#if CMD_LED > 0
// data LED hang-time (makes it more obvious to user that data was rec'd)
static const int DataLedDelay = 250;
// last time data was rec'd
static unsigned long lastRxData = 0;
// cmd LED state
static uint8_t lastCmdLED = LOW;
#endif

// last CW transition
static unsigned long lastTime = 0;

// used to detect end of transmission
static bool hanging = false;

// echo USB RX data back to host
static bool echo = false;

// currently parsing a command
static bool cmd = false;

// TX mode (otherwise RX)
static bool tx = false;

// tracks PTT state without reading PTT pin
static bool pttState = false;

// TX element timing
static unsigned long curTxStart = 0;
static unsigned long lastKeyDown = 0;
static unsigned curTxLength = 0;
static CwElement cwOut;

// PTT timing
static unsigned pttLeadIn = 100;			// msec
static unsigned pttLeadOut = 250;			// msec

// Limits
static const int MaxPttLeadIn = 500;		// msec
static const int MaxPttLeadOut = 1000;		// msec
static const int MaxWPM = 60;				// WPM
static const int MinWPM = 5;				// WPM
static const unsigned long
	MaxBounceInterval = 1200 / MaxWPM;		// msec
static const unsigned MinBoxCarLength = 8;		// elements
static const unsigned MaxBoxCarLength = 65535;	// elements

static const int KeyDebounceInterval = 3;	// msec

static const int InputStringLength = 16;
static const int CommandStringLength = 16;

// control keys
static const char ESC = 0x1B;		// key used for all-stop
static const char CtlC = 0x03;		// alternative key for all-stop


////////////////////////////////////////////////////////////////
//
//  Support Methods
//
////////////////////////////////////////////////////////////////

#if CW_DATA_ANALOG == true
// stores the audio level for diagnostics
static int8_t audio_max = 0;

//
//  A/D Interrupt Handler -- called when new A/D value ready
//
ISR(ADC_vect) {
	// the A/D input var
	static int8_t sample = 0;

	//
	//  NOTE: BE VERY FAST HERE; this handler is being called
	//        thousands of times per second, and doing too
	//        much work here will cause samples to be lost.
	//

	// read the current A/D value and remove the DC offset
	sample = ADCH - ANALOG_DC_OFFSET;

	// check for clipping; store max level
	const uint8_t alevel = abs(sample);
	if (alevel > audio_max)
		audio_max = alevel;

	// do the demod
	audioState = ! detector.run(sample);
}
#endif

//
//  ProcessElement(...) - run the decoder chain on a state change
//
static void ProcessElement(unsigned long now, unsigned width, bool state) {
	if (width > 1000) {
		width = 1000;
	}
	CwElement cw;
	cw.Mark = (bool)(!state);	// the detector pulls the pin LOW on contact (567, button, relay, etc.)
	cw.Length = width;
	CwBuffer.Add(cw);			// TODO: check this for failure, to detect overrun
	
	// push the pulse into the timer
	if (Timing->Decode(CwBuffer, ElementBuffer)) {
		// if the timer produced elements, run those through the decoder
		byte ct = Decoder->Decode(ElementBuffer, ioBuffer, 8);
		// if that produced printable characters, print them
		if (ct > 0) {
			ioBuffer[ct] = 0;
			HostSerial.print(ioBuffer);
		}
	}
}


#ifdef VERBOSE_DEBUG
//
//  DumpElementBuffer() - used only for development debugging
//
static void DumpElementBuffer() {
	HostSerial.println(strcpy_P(ioBuffer, PSTR("Decoder->Encode(): OK")));
	for (int k = 0; k != ElementBuffer.Count(); ++k) {
		switch(ElementBuffer.ItemAt(k)) {
			case Dot: HostSerial.print('.'); break;
			case Dash: HostSerial.print('-'); break;
			case WordSpace:
			case DashSpace: HostSerial.print(' '); break;
		}
	}
}
#endif


//
//  setup()
//
void setup() {
	// start the serial port
	HostSerial.begin(SerialRate);

#ifdef VERBOSE_DEBUG
	HostSerial.println();
	HostSerial.print(strcpy_P(ioBuffer, PSTR("64 = ")));
	HostSerial.println(sizeof(uint64_t));
	HostSerial.print(strcpy_P(ioBuffer, PSTR("N  = ")));
	HostSerial.println(detector.N);
	HostSerial.print(strcpy_P(ioBuffer, PSTR("w  = ")));
	HostSerial.println(detector.dist);
	HostSerial.print(strcpy_P(ioBuffer, PSTR("fr = ")));
	HostSerial.println(detector.frac.toDouble());
	HostSerial.print(strcpy_P(ioBuffer, PSTR("1m = ")));
	HostSerial.println(detector.frac_1m.toDouble());
	HostSerial.print(strcpy_P(ioBuffer, PSTR("fs = ")));
	HostSerial.println(GetFs(CPU_FREQ, SAMPLE_RATE_DIVISOR));
#endif
	
	// allocate strings
	cmdText.reserve(CommandStringLength);
	inputString.reserve(InputStringLength);
	inputString = strcpy_P(ioBuffer, PSTR("#VERSION;"));
	inputReady = true;

	// the 'command' LED
	#if CMD_LED > 0
	pinMode(CMD_LED, OUTPUT);
	digitalWrite(CMD_LED, LOW);
	lastCmdLED = LOW;

	// send 'K' to the command LED (just a toy)
	digitalWrite(CMD_LED, HIGH);
	delay(150);
	digitalWrite(CMD_LED, LOW);
	delay(50);
	digitalWrite(CMD_LED, HIGH);
	delay(50);
	digitalWrite(CMD_LED, LOW);
	delay(50);
	digitalWrite(CMD_LED, HIGH);
	delay(150);
	digitalWrite(CMD_LED, LOW);
	delay(250);
	#endif // CMD_LED > 0
	
	//
	// CW RX input
	//
#if CW_DATA_ANALOG == true
	// configure the A/D for continuous sampling
	ConfigureAnalogSampling(SAMPLE_RATE_DIVISOR);
#else
	// configure the RX pin for digial input from a demod IC
	pinMode(CW_RX_PIN, INPUT_PULLUP);
	KeyInput.attach(CW_RX_PIN);
	KeyInput.interval(KeyDebounceInterval);
#endif

	//
	// CW Decode LED
	//
#if CW_LED_PIN > 0
	pinMode(CW_LED_PIN, OUTPUT);
	digitalWrite(CW_LED_PIN, LOW);
#endif
	
	// CW TX (key) pin
	pinMode(CW_KEY_PIN, OUTPUT);
	digitalWrite(CW_KEY_PIN, LOW);

	// CW PTT pin
#if CW_PTT_PIN > 0
	pinMode(CW_PTT_PIN, OUTPUT);
	digitalWrite(CW_PTT_PIN, LOW);
#endif
	pttState = false;

	// save initial timestamps
	lastTime = millis();
#if CMD_LED > 0
	lastRxData = lastTime;
#endif
	
	// initialize WPM and set speed to auto-track
	Timing->RxWPM(25);
	Timing->TxWPM(25);
	Timing->RxMode(SpeedAuto);
	Timing->TxMode(SpeedAuto);
}


//
//  loop()
//
void loop() {
	// read the clock
	unsigned long now = millis();

	// extinguish the data LED
	#if CMD_LED > 0
	if ((lastCmdLED == HIGH) && (Elapsed(now, lastRxData) > DataLedDelay)) {
		lastCmdLED = LOW;
		digitalWrite(CMD_LED, LOW);
	}
	#endif
	
	//
	//  TX: if currently in transmit mode, service the TX queue
	//
	if (tx) {
		// if the current element has been sent...
		if (Elapsed(now, curTxStart) >= curTxLength) {
			// if there are no more elements to send, generate some more
			if (CwBuffer.Count() == 0 && TxBuffer.Count()) {
				char next;
				if (TxBuffer.Remove(next) && Decoder->Encode(ElementBuffer, &next, 1)) {
					#ifdef VERBOSE_DEBUG
					DumpElementBuffer();
					#endif

					// encode the char
					Timing->Encode(ElementBuffer, CwBuffer);
				}
			}
			
			// if there are elements to send, start sending the first one
			if (CwBuffer.Remove(cwOut)) {
				curTxStart = now;
				curTxLength = cwOut.Length;
				digitalWrite(CW_KEY_PIN, cwOut.Mark);
				if (!cwOut.Mark) {
					lastKeyDown = now;
				}
							
				#ifdef VERBOSE_DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("New Element: M=")));
				HostSerial.print(cwOut.Mark);
				HostSerial.print(strcpy_P(ioBuffer, PSTR("; L=")));
				HostSerial.println(cwOut.Length);
				#endif
			} else {
				tx = false;
#if CW_DATA_ANALOG == true
				// turn the receiver back on
				SetSamplingStatus(true);
#endif
			}
		}
	} // if (tx)...

	// key-up (semi break-in timer expired)
	now = millis();
	if (!tx && pttState && Elapsed(now, lastKeyDown) > pttLeadOut) {
#if CW_PTT_PIN > 0
		digitalWrite(CW_PTT_PIN, LOW);
#endif
		pttState = false;
	}
	
	// if terminal/serial/USB user input ready
	if (inputReady) {
		// light the CMD LED
		#if CMD_LED > 0
		lastRxData = millis();
		if (lastCmdLED == LOW) {
			lastCmdLED = HIGH;
			digitalWrite(CMD_LED, HIGH);
		}
		#endif
		
		// for each char received from the USB...
		for (int i = 0; i != inputString.length(); ++i) {
			char ch = inputString.charAt(i);
			
			//
			//  CMD: if currently in command mode...
			//
			if (cmd) {
				if (ch == '#') {
					// if a '#' sent mid-command, start over
					cmdText = "";
				} else if (ch == ';') {
					// terminate and process the command string
					cmd = false;
					bool ok = false;
					if (echo) HostSerial.print(';');

					// parse the command and argument
					SplitCommand(cmdText, command, argument);
					if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOC")))) {
						//
						//  QUERY DOC
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += strcpy_P(ioBuffer, PSTR(":RXWPM,TXWPM,RXMODE,TXMODE,"));
							cmdText += strcpy_P(ioBuffer, PSTR("PTTIN,PTTOUT,ECHO,DOT,BOUNCE,"));
							cmdText += strcpy_P(ioBuffer, PSTR("BCL,DZ,VERSION"));
						} else {
							//
							//  TODO: process individual query commands
							//
							ok = false;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("RXWPM")))) {
						//
						//  QUERY/SET WPM (RX)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(Timing->RxWPM());
						} else {
							int newWPM = argument.toInt();
							if (newWPM >= MinWPM && newWPM <= MaxWPM) {
								Timing->RxWPM(newWPM);
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TXWPM")))) {
						//
						//  QUERY/SET WPM (TX)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(Timing->TxWPM());
						} else {
							int newWPM = argument.toInt();
							if (newWPM >= MinWPM && newWPM <= MaxWPM) {
								Timing->TxWPM(newWPM);
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ECHO")))) {
						//
						//  QUERY/SET ECHO
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += echo ? "1" : "0";
						} else {
							int newEcho = argument.toInt();
							echo = newEcho ? true : false;
							ok = true;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTIN")))) {
						//
						//  QUERY/SET PTT LEAD-IN
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(pttLeadIn);
						} else {
							int newPTT = argument.toInt();
							if (newPTT >= 0 && newPTT <= MaxPttLeadIn) {
								pttLeadIn = newPTT;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTOUT")))) {
						//
						//  QUERY/SET PTT LEAD-OUT
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(pttLeadOut);
						} else {
							int newPTT = argument.toInt();
							if (newPTT >= 0 && newPTT <= MaxPttLeadOut) {
								pttLeadOut = newPTT;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOT")))) {
						//
						//  QUERY DOT LENGTH
						//
						ok = true;
						cmdText += '=';
						cmdText += (int)(Timing->DotLength());
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("GAP")))) {
						//
						//  QUERY SAFETY GAP
						//
						ok = true;
						cmdText += '=';
						cmdText += (int)(Timing->SafetyGap());
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("RXMODE")))) {
						//
						//  ENABLE/DISABLE RX WPM AUTO-TRACK
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Timing->RxMode() == SpeedAuto ? "AUTO" : "MANUAL";
						} else {
							if (argument.equalsIgnoreCase("AUTO")) {
								ok = true;
								Timing->RxMode(SpeedAuto);
							} else if (argument.equalsIgnoreCase("MANUAL")) {
								ok = true;
								Timing->RxMode(SpeedManual);
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TXMODE")))) {
						//
						//  ENABLE/DISABLE TX WPM AUTO-TRACK
						//
						ok = true;
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Timing->TxMode() == SpeedAuto ? "AUTO" : "MANUAL";
						} else {
							if (argument.equalsIgnoreCase("AUTO")) {
								ok = true;
								Timing->TxMode(SpeedAuto);
							} else if (argument.equalsIgnoreCase("MANUAL")) {
								ok = true;
								Timing->TxMode(SpeedManual);
							}
						}
#if CW_DATA_ANALOG == true
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DL")))) {
						// return the current detector internal level
						ok = true;
						cmdText += '=';
						cmdText += detector.level();
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DS")))) {
						// return the current detector output
						ok = true;
						cmdText += '=';
						cmdText += audioState ? strcpy_P(ioBuffer, PSTR("HIGH")) : strcpy_P(ioBuffer, PSTR("LOW"));
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVELMAX")))) {
						// return the highest volume level that can be returned
						ok = true;
						cmdText += '=';
						cmdText += 128;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVEL")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText += '=';
						cmdText += audio_max;
						audio_max = 0;
#else
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("BOUNCE")))) {
						//
						//  GET/SET CW INPUT DEBOUNCE INTERVAL
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(KeyInput.interval());
						} else {
							int newBounce = argument.toInt();
							if (newBounce >= 0 && newBounce <= MaxBounceInterval) {
								KeyInput.interval(newBounce);
								ok = true;
							}
						}
#endif
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("BCL")))) {
						//
						//  GET/SET CW SPEED TRACKING BOXCAR SIZE
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(Timing->BoxCarLength());
						} else {
							unsigned newBCL = argument.toInt();
							if (newBCL >= MinBoxCarLength && newBCL <= MaxBoxCarLength) {
								Timing->BoxCarLength(newBCL);
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DZ")))) {
						//
						//  GET/SET CW SPEED TRACKING DEAD ZONE (percent)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += (int)(Timing->MinimumAverageDistance() * 100);
						} else {
							unsigned newDZ = argument.toInt();
							if (newDZ >= 0 && newDZ <= 100) {
								Timing->MinimumAverageDistance((float)newDZ / 100.0);
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("VERSION")))) {
						//
						//  QUERY VERSION NUMBER
						//
						ok = true;
						cmdText += '=';
						cmdText += strcpy_P(ioBuffer, PSTR(VERSION_STRING));
					}

					//
					//  TODO: process future command contents
					//

					// respond with #OK or #ERR
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					HostSerial.print(ok ? strcpy_P(ioBuffer, PSTR("#OK:")) : strcpy_P(ioBuffer, PSTR("#ERR:")));
					HostSerial.print(cmdText);
					HostSerial.print(';');
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					cmdText = "";
				} else {
					if (isAlpha(ch) || isDigit(ch) || isPunct(ch) || (ch == ' ')) {
						// add the char to the command string
						if (cmdText.length() < CommandStringLength)
							cmdText += ch;
						
						// echo the char, if enabled
						if (echo) {
							HostSerial.print(ch);
						}
					}
				}
				continue;
			} // if (cmd)....
			
			// start command mode??
			if ((!cmd) && (ch == '#')) {
				cmd = true;
				if (echo) HostSerial.print('#');
				continue;
			}

			// user hard-stop?
			if (ch == ESC || ch == CtlC) {
				digitalWrite(CW_KEY_PIN, LOW);
#if CW_PTT_PIN > 0
				digitalWrite(CW_PTT_PIN, LOW);
#endif
				TxBuffer.Clear();
				CwBuffer.Clear();
				ElementBuffer.Clear();
				pttState = false;
				tx = false;
#if CW_DATA_ANALOG == true
				// turn the receiver back on
				SetSamplingStatus(true);
#endif
				continue;
			}

			// echo the char, if enabled
			if (echo) {
				HostSerial.print(ch);
				if (ch == '\r') {
					HostSerial.print('\n');
				}
			}

			// upper-case the char to search for a TX pattern
			ch = toupper(ch);

			// switch to TX mode, if needed
			if (!tx) {
				CwBuffer.Clear();
				ElementBuffer.Clear();
			}
			
			#ifdef VERBOSE_DEBUG
			HostSerial.print(strcpy(ioBuffer, PSTR("Send: ")));
			HostSerial.println(ch);
			#endif

			// if we are already in TX mode, just queue up the character
			if (tx) {
				TxBuffer.Add(ch);
			} else if (Decoder->Encode(ElementBuffer, &ch, 1)) {
				#ifdef VERBOSE_DEBUG
				DumpElementBuffer();
				#endif
				
				// otherwise, generate characters, and go into TX mode
				if (Timing->Encode(ElementBuffer, CwBuffer)) {
					if (CwBuffer.Remove(cwOut)) {
#if CW_DATA_ANALOG == true
						// turn the receiver off
						SetSamplingStatus(false);
#endif
						// PTT lead-in
#if CW_PTT_PIN > 0
						digitalWrite(CW_PTT_PIN, HIGH);
						delay(pttLeadIn);
#endif
						pttState = true;

						// start sending
						tx = true;
						curTxStart = millis();
						curTxLength = cwOut.Length;
						digitalWrite(CW_KEY_PIN, cwOut.Mark);
						if (!cwOut.Mark) {
							lastKeyDown = curTxStart;
						}
						
						#ifdef VERBOSE_DEBUG
						HostSerial.print(strcpy_P(ioBuffer, PSTR("New Element: M=")));
						HostSerial.print(cwOut.Mark);
						HostSerial.print(strcpy_P(ioBuffer, PSTR("; L=")));
						HostSerial.println(cwOut.Length);
						#endif
					}
				}
			} // !tx
		}

done:
		// clear the string:
		inputString = "";
		inputReady = false;
	}

	//
	//   RX: read data from input pin
	//
	if (!tx) {
		// if the state has changed
#if CW_DATA_ANALOG == true
		if (audioState != lastState) {
			bool newState = lastState;
#else
		if (KeyInput.update()) {
			bool newState = KeyInput.read();
#endif
			now = millis();
			
			// calculate the pulse width
			unsigned long width = Elapsed(now, lastTime);

#if CW_LED_PIN != 0
#if CW_DATA_ANALOG == true
			digitalWrite(CW_LED_PIN, newState);
#else
			digitalWrite(CW_LED_PIN, !newState);
#endif
#endif
			// update the decoder
			ProcessElement(now, width, lastState);
			
			// if the new state is key-up, look for trailing space
			if (newState) {
				hanging = true;
			}

			lastTime = now;		
#if CW_DATA_ANALOG == true
			lastState = audioState;
#else
			lastState = newState;
#endif
		} else {
			// handling the trailing edge of the last key-down element
			if (lastState && hanging) {
				now = millis();
				unsigned ws = Timing->DotLength() * Timing->MinimumWordSpace();
				if (Elapsed(now, lastTime) > ws) {
					ProcessElement(now, ws, true);
					hanging = false;
				}
			}
		}
	} // if !tx

	// service the serial port
	while (HostSerial.available() && (inputString.length() < InputStringLength)) {
		// get the new byte:
		char inChar = (char)HostSerial.read();
		// add it to the inputString:
		inputString += inChar;
		inputReady = true;
	}
}

// EOF
